# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections.abc import Callable
from functools import wraps
from pathlib import Path
from typing import Any

import requests

from lightworks_remote.utils import AuthenticationError
from lightworks_remote.utils.configuration import APIURL


def catch_error_codes(
    func: Callable[..., requests.Response],
) -> requests.Response:
    """
    Check for generic errors and raise an exception for these.
    """

    @wraps(func)
    def run_and_catch(
        self: "HTTPInterface", *args: Any, **kwargs: Any
    ) -> requests.Response:
        response = func(self, *args, **kwargs)
        if response.status_code == 401:
            raise AuthenticationError(
                "Unable to authenticate with job submission service, this may "
                "result from an invalid or expired access token being provided."
            )
        if response.status_code == 403:
            raise AuthenticationError(
                "Not authenticated to access the requested resource."
            )
        return response

    return run_and_catch


TIMEOUT = 10  # Set timeout for requests


class HTTPInterface:
    """
    Generic http interface to automatically include base url and certificate
    data into http requests.

    Args:

        url (APIURL) : The APIURl object which details the base url.

        certificate (str) : A file path to provide a certificate for
            authentication with the http endpoint.

    """

    def __init__(self, url: APIURL, certificate: str | Path) -> None:
        self._url = url
        self._certificate = str(certificate)

    @catch_error_codes
    def _get(self, loc: str, **kwargs: Any) -> requests.Response:
        return self._get_no_catch(loc, **kwargs)

    @catch_error_codes
    def _post(self, loc: str, **kwargs: Any) -> requests.Response:
        return self._post_no_catch(loc, **kwargs)

    @catch_error_codes
    def _put(self, loc: str, **kwargs: Any) -> requests.Response:
        return self._put_no_catch(loc, **kwargs)

    def _get_no_catch(self, loc: str, **kwargs: Any) -> requests.Response:
        return requests.get(
            self._url.url + loc,
            verify=self._certificate,
            timeout=TIMEOUT,
            **kwargs,
        )

    def _post_no_catch(self, loc: str, **kwargs: Any) -> requests.Response:
        return requests.post(
            self._url.url + loc,
            verify=self._certificate,
            timeout=TIMEOUT,
            **kwargs,
        )

    def _put_no_catch(self, loc: str, **kwargs: Any) -> requests.Response:
        return requests.put(
            self._url.url + loc,
            verify=self._certificate,
            timeout=TIMEOUT,
            **kwargs,
        )



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass

from .qpu_details import QPUDetails

ARTEMIS_QPU_DETAILS_MAP = {
    "id": "id",
    "name": "name",
    "numberOfModes": "n_modes",
    "maxPhotonInputNumber": "max_photon_input",
    "maxDetectionFilter": "max_detection_filter",
    "defaultMaxSamples": "default_max_samples",
    "maxSamples": "max_samples",
    "isAvailable": "available",
    "lastUpdated": "last_updated",
}


@dataclass(kw_only=True)
class ArtemisQPU(QPUDetails):
    """
    Dataclass for storing all details of an artemis QPU.
    """

    name: str
    id: str
    n_modes: int
    max_photon_input: int
    max_detection_filter: int
    default_max_samples: int
    max_samples: dict[int, int]
    available: bool
    last_updated: str



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import NoneType
from typing import Any

from lightworks import PostSelection, State
from lightworks.sdk.tasks import SamplerTask
from lightworks.sdk.utils.heralding import add_heralds_to_state
from lightworks.sdk.utils.post_selection import DefaultPostSelection


class SamplerCompiler:
    """
    Creates and verifies a payload for performing sampling experiments on a
    remote QPU.

    Args:

        data (SamplerTask) : The sampler data to run on the QPU.

    """

    def __init__(self, data: SamplerTask) -> None:
        self.data = self._validate(data)

    def _validate(self, data: SamplerTask) -> SamplerTask:
        # Circuit
        if data.circuit.loss_modes != 0:
            raise ValueError(
                "Circuits to be executed on remote QPU cannot have any loss "
                "elements."
            )
        heralds = data.circuit.heralds
        for d in [heralds.input, heralds.output]:
            if d and max(d.values()) > 1:
                raise ValueError(
                    "Circuit to sampler from cannot have more than 1 heralded "
                    "photons per mode."
                )
        # Input
        for i in data.input_state:
            if i > 1:
                raise ValueError(
                    "Inputs must have a max of 1 photons per mode."
                )
        if sum(data.input_state) < 1:
            raise ValueError("Input state should contain at least one photon.")
        # Min photon detection
        if data.min_detection is not None:
            if data.min_detection < 1:
                raise ValueError(
                    "Minimum photon detection should be at least 1."
                )
            if data.min_detection > sum(data.input_state):
                raise ValueError(
                    "Minimum photon detection cannot be larger than input "
                    "photon number."
                )
        # Post-selection
        if not isinstance(
            data.post_selection, PostSelection | NoneType | DefaultPostSelection
        ):
            raise TypeError(
                "post_selection should be PostSelection object or None. "
                "Post-selection functions are not supported on remote QPUs."
            )
        # Source
        if data.source is not None:
            raise TypeError(
                "Custom source properties are not supported on remote QPUs."
            )
        # Detector
        if data.detector is not None:
            raise TypeError(
                "Custom detector properties are not supported on remote QPUs."
            )
        return data

    @property
    def full_input_state(self) -> State:
        """Returns the full input state including heralded modes."""
        return State(
            add_heralds_to_state(
                self.data.input_state, self.data.circuit.heralds.input
            )
        )

    def _generate_payload(
        self, direct_encoding: bool = False
    ) -> dict[str, Any]:
        """Used to create the payload for submission."""
        # Generate min detection number based on provided value and heralds
        min_detection = (
            sum(self.data.input_state)
            if self.data.min_detection is None
            else self.data.min_detection
        ) + sum(self.data.circuit.heralds.input.values())
        # Get post-selection data
        post_selection = self._get_post_selection()
        # Create dictionary and then return
        payload = {
            "n_modes": self.data.circuit.n_modes,
            "input": self.full_input_state,
            "n_samples": self.data.n_samples,
            "min_detection": min_detection,
            "direct_implementation": direct_encoding,
            "unitary": self.data.circuit.U_full,
        }
        if direct_encoding:
            payload["circuit_spec"] = self.data.circuit._circuit_spec
        # Only add post_selection if it is required
        if post_selection["post_select"] or post_selection["herald"]:
            payload["job_data"] = dict(post_selection)
        return payload

    def _get_post_selection(
        self,
    ) -> dict[str, list[tuple[tuple[int, ...], tuple[int, ...]]]]:
        """Generates required post-selection data for a particular payload"""
        post_selection = []
        # Get data from manually set post-selection
        if self.data.post_selection is not None and not isinstance(
            self.data.post_selection, DefaultPostSelection
        ):
            if not isinstance(self.data.post_selection, PostSelection):
                raise TypeError(
                    "Post-selection cannot be a function when executing "
                    "remotely, must be a lightworks PostSelection object."
                )
            herald_modes = list(self.data.circuit.heralds.output.keys())
            for rule in self.data.post_selection.rules:
                mapped_modes = [_map_mode(m, herald_modes) for m in rule.modes]
                if max(mapped_modes) >= self.data.circuit.n_modes:
                    raise ValueError(
                        "Provided post-selection rules detected to exceed size "
                        "of circuit."
                    )
                post_selection.append((tuple(mapped_modes), rule.n_photons))
        # And from heralds at the output
        herald: list[tuple[tuple[int, ...], tuple[int, ...]]] = []
        for m, n in self.data.circuit.heralds.output.items():
            herald.append(((m,), (n,)))
        return {"post_select": post_selection, "herald": herald}


def _map_mode(mode: int, herald_modes: list[int]) -> int:
    """
    Remaps a provided mode around the heralded modes of a circuit.
    """
    for i in sorted(herald_modes):
        if mode >= i:
            mode += 1
    return mode



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum


class Status(Enum):
    """
    Enumerate all possible job statuses.
    """

    ACCEPTED = "Accepted"
    COMPILING = "Compiling"
    SCHEDULED = "Scheduled"
    RUNNING = "Running"
    COMPLETED = "Completed"
    FAILED = "Failed"
    CANCELLED = "Cancelled"
    TIMEDOUT = "Timed Out"

    def __str__(self) -> str:
        """Overwrites str to return value instead of key."""
        return self.value

    @property
    def is_complete(self) -> bool:
        """Returns whether current status can be considered complete."""
        return self in {  # type: ignore [comparison-overlap]
            self.COMPLETED,
            self.FAILED,
            self.CANCELLED,
            self.TIMEDOUT,
        }

    @property
    def is_cancellable(self) -> bool:
        """Returns whether job is cancellable in current status."""
        return self in {  # type: ignore [comparison-overlap]
            self.ACCEPTED,
            self.COMPILING,
            self.SCHEDULED,
            self.RUNNING,
        }

    @property
    def has_results(self) -> bool:
        """Returns if job was able to generate a set of results."""
        return self in {self.COMPLETED, self.TIMEDOUT}  # type: ignore [comparison-overlap]



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from types import NoneType

from requests import Response


class LightworksRemoteError(Exception):
    """General exception for lightworks remote module."""


class JobSubmissionError(LightworksRemoteError):
    """For errors arising during submission procedure."""


class AuthenticationError(LightworksRemoteError):
    """For errors relating to authentication of system."""


class ValidationError(LightworksRemoteError):
    """For errors arising from invalid values provided as part of a job."""


class TokenError(LightworksRemoteError):
    """For errors relating to getting/setting of authentication token."""


class ResultsError(LightworksRemoteError):
    """For errors relating to retrieval of results from the remote platform."""


class JobError(LightworksRemoteError):
    """General errors relating to a job."""


class JobNotFoundError(LightworksRemoteError):
    """For errors when a job with matching id is not found."""


class HTTPError(LightworksRemoteError):
    """For other errors relating to HTTP communication with platform."""

    def __init__(self, message: str, response: Response | None = None) -> None:
        if not isinstance(response, Response | NoneType):
            raise TypeError(
                "response should be a requests Response object or None."
            )
        if response is not None:
            message += f" (Code: {response.status_code} {response.reason})"
        super().__init__(message)



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from lightworks import (
    PhotonicCircuit,
    PostSelection,
    Sampler,
    State,
    Unitary,
    random_unitary,
)

from lightworks_remote.remote.sampler_compiler import SamplerCompiler


class TestSamplerTask:
    """Contains all unit tests relating to the SamplerTask."""

    def setup_method(self):
        """
        Assigns some circuit attribute and resets before each test.
        """
        self.circuit = PhotonicCircuit(6)
        self.input_state = State([1, 0, 0, 1, 0, 0])
        self.n_samples = 10000
        self.min_detection_number = 2
        self.post_selection = None
        self.all_args = {
            "circuit": self.circuit,
            "input_state": self.input_state,
            "n_samples": self.n_samples,
            "min_detection": self.min_detection_number,
            "post_selection": self.post_selection,
        }

    def test_basic_setup(self):
        """
        Sets up a SamplerTask to confirm basic initialisation functionality.
        """
        Sampler(**self.all_args)

    def test_basic_setup_only_req_args(self):
        """
        Sets up a SamplerTask to confirm basic initialisation functionality,
        providing only the required args and allowing others to assume default
        values.
        """
        Sampler(self.circuit, self.input_state, self.n_samples)

    def test_payload_contents(self):
        """
        Checks payload contents does not change.
        """
        task = Sampler(**self.all_args)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()

        expected_keys = [
            "n_modes",
            "input",
            "n_samples",
            "min_detection",
            "unitary",
            "direct_implementation",
            "circuit_spec",
        ]
        for k in payload:
            if k not in expected_keys:
                pytest.fail("Unexpected key present in payload.")

    def test_payload_values(self):
        """
        Checks that the values from the payload are the expected values for the
        provided job.
        """
        task = Sampler(**self.all_args)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()

        expected_values = {
            "n_modes": 6,
            "input": State([1, 0, 0, 1, 0, 0]),
            "n_samples": 10000,
            "min_detection": 2,
        }
        for k, v in expected_values.items():
            if k not in payload:
                pytest.fail(f"Key {k} not found in payload.")
            else:
                assert payload[k] == v

    def test_heralded_circuit(self):
        """
        Confirms that the correct payload values are generated when heralds are
        included within a circuit.
        """
        sub_circ = PhotonicCircuit(6)
        sub_circ.herald(0, 1)
        sub_circ.herald(5, 1)
        circ = PhotonicCircuit(6)
        circ.add(sub_circ, 1)
        # Generate task and get payload
        task = Sampler(circ, self.input_state, self.n_samples)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()
        # Then check against expected values
        expected_values = {
            "n_modes": 8,
            "input": State([1, 1, 0, 0, 1, 0, 1, 0]),
            "n_samples": 10000,
            "min_detection": 4,
        }
        for k, v in expected_values.items():
            if k not in payload:
                pytest.fail(f"Key {k} not found in payload.")
            else:
                assert payload[k] == v

    def test_lossy_circuit(self):
        """
        Checks that attempt to use a lossy circuit will raise an exception.
        """
        circuit = PhotonicCircuit(6)
        circuit.loss(0, 1)
        self.all_args["circuit"] = circuit
        task = Sampler(**self.all_args)
        with pytest.raises(ValueError):
            SamplerCompiler(task._generate_task())

    def test_no_min_detection(self):
        """
        Checks correct default value for min_detection is found in cases where
        this is not specified.
        """
        task = Sampler(self.circuit, self.input_state, self.n_samples)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()
        assert payload["min_detection"] == 2

    def test_heralds(self):
        """
        Checks that input state is set correctly when
        """
        circ = Unitary(random_unitary(6))
        circ.herald(1, 1)
        circ.herald(4, 1)
        task = Sampler(circ, State([1, 0, 0, 0]), 10000)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()
        assert payload["input"] == State([1, 1, 0, 0, 1, 0])

    def test_no_post_selection(self):
        """
        Checks that if no heralding/post-selection is set then the payload does
        not contain the post-selection field.
        """
        task = Sampler(**self.all_args)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()
        assert "post_selection" not in payload

    def test_add_post_selection(self):
        """
        Checks that when post-selection is added with the add_post_selection
        option then this is set over the correct modes.
        """
        p = PostSelection()
        p.add([0, 2], 1)
        self.all_args["post_selection"] = p
        task = Sampler(**self.all_args)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()

        assert payload["job_data"]["post_select"][0][0] == (0, 2)

    def test_add_post_selection_and_heralds(self):
        """
        Checks that when post-selection is added with the add_post_selection
        option then this is set over the correct modes.
        """
        circ = Unitary(random_unitary(6))
        circ.herald(1, 1)
        circ.herald(4, 1)

        p = PostSelection()
        p.add([0, 2], 1)

        task = Sampler(circ, State([1, 0, 0, 0]), 10000, post_selection=p)
        compiler = SamplerCompiler(task._generate_task())
        payload = compiler._generate_payload()

        assert payload["job_data"]["post_select"][0][0] == (0, 3)
        assert payload["job_data"]["herald"][0][0] == (1,)
        assert payload["job_data"]["herald"][1][0] == (4,)



# Copyright 2025 - 2025 Aegiq Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import asdict, dataclass, is_dataclass
from typing import Any

import numpy as np
from lightworks import State
from numpy.typing import NDArray


class Payload:
    """Generic Payload which allows for any fields to be specified as kwargs."""

    def __init__(self, **kwargs: Any) -> None:
        self.__all_attrs = []
        for k, v in kwargs.items():
            setattr(self, k, v)
            self.__all_attrs.append(k)

    def __str__(self) -> str:
        """String representation of payload."""
        data = [
            f"{k}={getattr(self, k)}"
            if not isinstance(k, str)
            else f"{k}='{getattr(self, k)}'"
            for k in self.__all_attrs
        ]
        joined_data = ", ".join(data)
        return f"Payload({joined_data})"

    def __repr__(self) -> str:
        """Representation of payload."""
        return str(self)

    @property
    def payload(self) -> dict[str, Any]:
        """Returns created payload for the Artemis system."""
        if is_dataclass(self):
            return asdict(self)
        return {k: getattr(self, k) for k in self.__all_attrs}


@dataclass(kw_only=True)
class ArtemisPayload(Payload):
    """
    Stores details of payload for Artemis.
    """

    qpu: str
    job_name: str
    lightworks_version: str
    n_modes: int
    input: State | list[int]
    n_samples: int
    min_detection: int
    direct_implementation: bool
    unitary: NDArray[np.complex128]
    circuit_spec: list[Any] | None = None
    job_data: dict[str, Any] | None = None



from setuptools import setup, find_packages

setup(
    name = "lightworks_remote",
    author = "Aegiq Ltd.",
    use_scm_version = True,
    description = "Remote job submission extension to lightworks.",
    license="Apache 2.0",
    packages = find_packages(where=".", exclude = ["tests"]),
    package_data={"lightworks_remote": ["py.typed"]},
    python_requires = ">=3.10",
    install_requires = [
        "lightworks>=2.3.0", 
        "requests>=2.32.3", 
        "pandas", 
        "multimethod>=1.11.2"
    ],
    entry_points={
        'console_scripts': [
            'lightworks_remote = lightworks_remote.__main__:main',
        ],
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Operating System :: OS Independent",
    ]
)
